/*
  MATH.INC - Mathematical graph drawer for POV-Ray 3.0+
  
  PLEASE read MATH.TXT before attempting to use this!

  - Sean O'Malley
    March, 1999
*/


// Check for neglected options.

#local badoption = 0;

// Whether to center the graph over the origin - not a required variable.
#ifndef (centerxy)
 #local centerxy = true;
#end

// Whether to center the average Z value to the origin. smooth_graph only.
#ifndef (centerz)
 #local centerz = false;
#end

#ifndef (domath)
 #macro domath(a,b)
  pow(e,-(mysqr(mydistance(0,0,a,b))/4))
 #end
 #local badoption = 1;
#end

#ifndef (minx)
 #declare minx = -1;
 #local badoption = 1;
#end

#ifndef (maxx)
 #declare maxx = 1;
 #local badoption = 1;
#end

#ifndef (miny)
 #declare miny = -1;
 #local badoption = 1;
#end

#ifndef (maxy)
 #declare maxy = 1;
 #local badoption = 1;
#end

#ifndef (graph_size)
 #declare graph_size = 15;
 #local badoption = 1;
#end

#ifndef (smooth_graph)
 #declare smooth_graph = false;
 #local badoption = 1;
#end

#if (badoption=1)
 #warning "\nWARNING!!! One or more settings were not defined for MATH.INC and default(s) were used"
 #warning "\ninstead.  Please check documentation for more information.\n\n"
#end

// Check if it is going to be extremely messed up.

#local badoption = 0;

#if (minx > maxx) #local badoption = 1; #end
#if (miny > maxy) #local badoption = 1; #end
#if (badoption=1)
 #warning "\nWARNING!!! Minimum/maximum values for X and/or Y are set incorrectly.  A graph will"
 #warning "\nprobably not be produced.\n\n"
#end

#local incrementx = (maxx-minx) / graph_size;
#local incrementy = (maxy-miny) / graph_size;

// Three unnecessary constants and macros, but useful.

#declare e = 2.71828182846;
#macro mydistance(x1,y1,x2,y2)
 vlength(<x2,y2,0>-<x1,y1,0>)
#end
#macro mysqr(a)
 (a*a)
#end

// This checks if a triangle is over the edge of the array,
// and if so it will return the nearest one to get an approximated
// vector to compute the normal at the edge.

#macro getvect(wherex,wherey)
 #local invent = 0; #local w1 = 0; #local w2 = 0;
 #if (wherex < 0)      #local invent = 1; #end
 #if (wherex > indexx) #local invent = 2; #end
 #if (wherey < 0)      #local invent = 3; #end
 #if (wherey > indexy) #local invent = 4; #end
 #switch (invent)
  #case (1) #local w1 = 1;  #break
  #case (2) #local w1 = -1; #break
  #case (3) #local w2 = 1;  #break
  #case (4) #local w2 = -1; #break
 #end
 vertices[wherex + w1][wherey + w2];
#end

#declare xx = minx;
#declare yy = miny;
// Find the (X,Y) center of the graph to shift the graph to the origin
// so it doesn't go all over the place - especially in animations!
#if (centerxy)
 #local graph_center = <(maxx+minx)/2,(maxy+miny)/2,0>;
#else
 #local graph_center = <0,0,0>;
#end

// Draw a recentered square in 3-D space.
#macro make_square (x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4)
 triangle { <x1,y1,z1>-graph_center,
            <x2,y2,z2>-graph_center,
            <x3,y3,z3>-graph_center }
 triangle { <x1,y1,z1>-graph_center,
            <x3,y3,z3>-graph_center,
            <x4,y4,z4>-graph_center }
#end

// Draw a recentered smooth_triangle.
#macro mysmooth_triangle(v1,n1,v2,n2,v3,n3)
 smooth_triangle
  { v1-graph_center, n1,
    v2-graph_center, n2,
    v3-graph_center, n3 }
#end

#if (smooth_graph)
 // So you want a smooth graph ...
 #local indexx = 0;
 #local indexy = 0;
 #declare graph_size = graph_size + 1;
 #local vertices = array[graph_size][graph_size]
 #local normals = array[graph_size][graph_size]
 #local zcount = 0; #local ztotal = 0;   // These assist in computing the average Z value for centerz.
 // Run through whole graph, saving values in the "vertices" array.
 #while (xx <= maxx)
  #while (yy <= maxy)
   #local zz = domath(xx,yy);
   #local vertices[indexx][indexy] = <xx,yy,zz>;
   #if (centerz)
    #local ztotal = ztotal + zz;
    #local zcount = zcount + 1;
   #end
   #declare yy = yy + incrementy;
   #local indexy = indexy + 1;
  #end  
  #local yy = miny;
  #declare xx = xx + incrementx;
  #local indexx = indexx + 1;
  #if (xx <= maxx) #local indexy = 0; #end
 #end
 #local indexx = indexx - 1;
 #local indexy = indexy - 1;
 #if (centerz)
  #local graph_center = graph_center + <0,0,ztotal/zcount>;
  #undef ztotal
  #undef zcount
 #end
 // Have an array of all points necessary, now find the average 
 // normal of the four (right) triangles that meet at that point
 // and save in the "normals" array.
 #local xx = 0;
 #local yy = 0;
 #while (xx <= indexx)
  #while (yy <= indexy)
   #local n1 = getvect(xx-1,yy)
   #local n2 = getvect(xx,yy+1)
   #local n3 = getvect(xx+1,yy)
   #local n4 = getvect(xx,yy-1)
   #local normals[xx][yy] = vnormalize((vcross(n1,n2)+vcross(n2,n3)+vcross(n3,n4)+vcross(n4,n1))/4);
   #local yy = yy + 1;
  #end
  #local yy = 0;
  #local xx = xx + 1; 
 #end
 // Finally, draw triangles.
 #declare math_object = mesh {
  #local xx = 0;
  #local yy = 0;
  #while (xx < indexx)
   #while (yy < indexy)
    // These two triangles form a square.
    mysmooth_triangle (
     vertices[xx][yy], normals[xx][yy],
     vertices[xx + 1][yy], normals[xx + 1][yy],
     vertices[xx + 1][yy + 1], normals[xx + 1][yy + 1]
    )
    mysmooth_triangle (
     vertices[xx][yy], normals[xx][yy],
     vertices[xx][yy + 1], normals[xx][yy + 1],
     vertices[xx + 1][yy + 1], normals[xx + 1][yy + 1]
    )
    #local yy = yy + 1;
   #end
   #local yy = 0;
   #local xx = xx + 1;
  #end 
 }
#else
 #declare math_object = mesh {
  // A non-smooth graph.  Piece of cake - no smoothing, normals, or nothin'.
  // Run through the graph, compute four points to form a square, and
  // make the square out of triangles.
  #while (xx <= (maxx - incrementx))
   #while (yy <= (maxy - incrementy))
    make_square(xx,yy,domath(xx,yy),
                xx+incrementx,yy,domath(xx+incrementx,yy),
                xx+incrementx,yy+incrementy,domath(xx+incrementx,yy+incrementy),
                xx,yy+incrementy,domath(xx,yy+incrementy))
    #declare yy = yy + incrementy;
   #end
   #declare yy = miny;
   #declare xx = xx + incrementx;
  #end
 }
#end
